@@ -3,49 +3,56 @@
 //!
 //! It does not do anything more than that, which makes it so fast.
 
-use std::error::Error;
-use std::io::{BufReader, BufWriter, Read, Write};
+use std::{
+    io,
+    io::{Read, Write},
+};
 
-///
 /// Set the indentation used for the formatting.
 ///
 /// Note: It is *not* recommended to set indentation to anything oder than some spaces or some tabs,
 /// but nothing is stopping you from doing that.
 #[derive(Debug, Copy, Clone, Eq, PartialEq, Hash)]
 pub enum Indentation<'a> {
-    /// Use the default indentation, which is two spaces
-    Default,
+    /// Fast path for two spaces
+    TwoSpace,
+    /// Fast path for four spaces
+    FourSpace,
+    /// Fast path for tab
+    Tab,
     /// Use a custom indentation String
     Custom(&'a str),
 }
 
-///
-/// # Formats a json string
-///
-/// The indentation can be set to any value using [Indentation](jsonformat::Indentation)
-/// The default value is two spaces
-/// The default indentation is faster than a custom one
-///
-pub fn format_json(json: &str, indentation: Indentation) -> String {
-    let mut reader = BufReader::new(json.as_bytes());
-    let mut writer = BufWriter::new(Vec::new());
-
-    format_json_buffered(&mut reader, &mut writer, indentation).unwrap();
-    String::from_utf8(writer.into_inner().unwrap()).unwrap()
+impl Default for Indentation<'_> {
+    fn default() -> Self {
+        Self::TwoSpace
+    }
 }
 
-///
 /// # Formats a json string
 ///
-/// The indentation can be set to any value using [Indentation](jsonformat::Indentation)
+/// The indentation can be set to any value using [`Indentation`]
 /// The default value is two spaces
 /// The default indentation is faster than a custom one
+pub fn format(json: &str, indentation: Indentation) -> String {
+    let mut reader = json.as_bytes();
+    let mut writer = Vec::with_capacity(json.len());
+
+    format_reader_writer(&mut reader, &mut writer, indentation).unwrap();
+    String::from_utf8(writer).unwrap()
+}
+
+/// # Formats a json string
 ///
-pub fn format_json_buffered<R, W>(
-    reader: &mut BufReader<R>,
-    writer: &mut BufWriter<W>,
+/// The indentation can be set to any value using [`Indentation`]
+/// The default value is two spaces
+/// The default indentation is faster than a custom one
+pub fn format_reader_writer<R, W>(
+    reader: R,
+    mut writer: W,
     indentation: Indentation,
-) -> Result<(), Box<dyn Error>>
+) -> io::Result<()>
 where
     R: Read,
     W: Write,
@@ -94,8 +101,8 @@ where
                     indent_level = indent_level.saturating_sub(1);
                     if !newline_requested {
                         // see comment below about newline_requested
-                        writer.write_all(&[b'\n'])?;
-                        indent_buffered(writer, indent_level, indentation)?;
+                        writer.write_all(b"\n")?;
+                        indent(&mut writer, indent_level, indentation)?;
                     }
                 }
                 b':' => {
@@ -112,8 +119,8 @@ where
                 // newline only happens after { [ and ,
                 // this means we can safely assume that it being followed up by } or ]
                 // means an empty object/array
-                writer.write_all(&[b'\n'])?;
-                indent_buffered(writer, old_level, indentation)?;
+                writer.write_all(b"\n")?;
+                indent(&mut writer, old_level, indentation)?;
             }
 
             if auto_push {
@@ -124,22 +131,27 @@ where
         }
     }
 
+    // trailing newline
+    writer.write_all(b"\n")?;
+
     Ok(())
 }
 
-fn indent_buffered<W>(
-    writer: &mut BufWriter<W>,
-    level: usize,
-    indent_str: Indentation,
-) -> Result<(), Box<dyn Error>>
+fn indent<W>(writer: &mut W, level: usize, indent_str: Indentation) -> io::Result<()>
 where
-    W: std::io::Write,
+    W: Write,
 {
     for _ in 0..level {
         match indent_str {
-            Indentation::Default => {
+            Indentation::TwoSpace => {
                 writer.write_all(b"  ")?;
             }
+            Indentation::FourSpace => {
+                writer.write_all(b"    ")?;
+            }
+            Indentation::Tab => {
+                writer.write_all(b"\t")?;
+            }
             Indentation::Custom(indent) => {
                 writer.write_all(indent.as_bytes())?;
             }
@@ -155,28 +167,28 @@ mod test {
 
     #[test]
     fn echoes_primitive() {
-        let json = "1.35";
-        assert_eq!(json, format_json(json, Indentation::Default));
+        let json = "1.35\n";
+        assert_eq!(json, format(json, Indentation::TwoSpace));
     }
 
     #[test]
     fn ignore_whitespace_in_string() {
-        let json = "\" hallo \"";
-        assert_eq!(json, format_json(json, Indentation::Default));
+        let json = "\" hallo \"\n";
+        assert_eq!(json, format(json, Indentation::TwoSpace));
     }
 
     #[test]
     fn remove_leading_whitespace() {
         let json = "   0";
-        let expected = "0";
-        assert_eq!(expected, format_json(json, Indentation::Default));
+        let expected = "0\n";
+        assert_eq!(expected, format(json, Indentation::TwoSpace));
     }
 
     #[test]
     fn handle_escaped_strings() {
         let json = "  \" hallo \\\" \" ";
-        let expected = "\" hallo \\\" \"";
-        assert_eq!(expected, format_json(json, Indentation::Default));
+        let expected = "\" hallo \\\" \"\n";
+        assert_eq!(expected, format(json, Indentation::TwoSpace));
     }
 
     #[test]
@@ -184,8 +196,9 @@ mod test {
         let json = "{\"a\":0}";
         let expected = "{
   \"a\": 0
-}";
-        assert_eq!(expected, format_json(json, Indentation::Default));
+}
+";
+        assert_eq!(expected, format(json, Indentation::TwoSpace));
     }
 
     #[test]
@@ -195,8 +208,9 @@ mod test {
   1,
   2,
   null
-]";
-        assert_eq!(expected, format_json(json, Indentation::Default));
+]
+";
+        assert_eq!(expected, format(json, Indentation::TwoSpace));
     }
 
     #[test]
@@ -210,9 +224,10 @@ mod test {
   {
     \"a\": null
   }
-]";
+]
+";
 
-        assert_eq!(expected, format_json(json, Indentation::Default));
+        assert_eq!(expected, format(json, Indentation::TwoSpace));
     }
 
     #[test]
@@ -225,8 +240,9 @@ mod test {
   {
     \"a\": null
   }
-]";
+]
+";
 
-        assert_eq!(expected, format_json(expected, Indentation::Default));
+        assert_eq!(expected, format(expected, Indentation::TwoSpace));
     }
 }
